#define _GNU_SOURCE
#include <stdio.h>
#include <assert.h>
#include <stdint.h>
#include <inttypes.h>
#include <libaio.h>
#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/callback.h>
#include <caml/signals.h>
#include <caml/custom.h>

#define BLOCKSIZE 4096
#define PAGESIZE 4096
#define AIO_MAX_IO 32

typedef struct Context {
  io_context_t ctx;
  int max_ios;
  int pending;
  struct iocb *iocbs[0];
} Context;

CAMLprim value caml_aio_run(value context);

void caml_aio_context_finalize(value v) {
  Context *ctx = (Context*)Data_custom_val(v);
  fprintf(stderr, "### caml_aio_context_finalize()\n");
  if (ctx->pending > 0) {
    fprintf(stderr, "Error: pending io on Aio.context.\n");
    fflush(stderr);
  }
  // FIXME: Can't throw exception. What's to do?
  assert(io_queue_release(ctx->ctx) == 0);
}

static struct custom_operations caml_aio_context_ops = {
  "vonbrederlow.de.aio.context",
  caml_aio_context_finalize,
  custom_compare_default,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};

/* context: fun max_ios -> context
external context: int -> context = "caml_aio_context"
*/
CAMLprim value caml_aio_context(value ml_max_ios) {
  CAMLparam1(ml_max_ios);
  CAMLlocal2(ml_ctx, ml_context);
  int max_ios = Int_val(ml_max_ios);
  intptr_t i;
  fprintf(stderr, "### caml_aio_context(%d)\n", max_ios);

  /*
   * context
   * callback
   * ...
   */
  ml_context = caml_alloc_custom(&caml_aio_context_ops, sizeof(Context) + 2 * max_ios * sizeof(struct iocb*), 0, 1);
  ml_ctx = caml_alloc_tuple(2 * max_ios + 1);
  Store_field(ml_ctx, 0, ml_context);
  for(i = 1; i <= 2 * max_ios; ++i) {
    Store_field(ml_ctx, i, Val_unit);
  }
  Context *context = (Context*)Data_custom_val(ml_context);
  memset(context, 0, sizeof(Context) + max_ios * sizeof(struct iocb*));
  for(i = 0; i < max_ios; ++i) {
    context->iocbs[i] = malloc(sizeof(struct iocb));
    // FIXME: throw exception
    assert(context->iocbs[i]);
    memset(context->iocbs[i], 0, sizeof(struct iocb));
    context->iocbs[i]->data = (void*)(2 * i + 1);
  }

  io_queue_init(max_ios, &context->ctx);
  context->max_ios = max_ios;
  context->pending = 0;

  CAMLreturn(ml_ctx);
}


typedef struct Buffer {
  char *buf;
  size_t len;
  size_t pos;
} Buffer;

void caml_aio_buffer_finalize(value v) {
  Buffer *buffer = (Buffer*)Data_custom_val(v);
  fprintf(stderr, "### caml_aio_buffer_finalize()\n");
  free(buffer->buf);
}

static struct custom_operations caml_aio_buffer_ops = {
  "vonbrederlow.de.aio.buffer",
  caml_aio_buffer_finalize,
  custom_compare_default,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};


/* read: fun ctx fd buf off fn -> ()
external read : context -> Unix.file_descr -> buffer -> int64 -> (unit -> unit) -> unit = "caml_aio_read"
*/
CAMLprim value caml_aio_read(value ml_ctx, value ml_fd, value ml_buffer, value ml_off, value ml_fn) {
  CAMLparam5(ml_ctx, ml_fd, ml_buffer, ml_off, ml_fn);
  fprintf(stderr, "### caml_aio_read()\n");
  Context *ctx = (Context*)Data_custom_val(Field(ml_ctx, 0));
  // FIXME: throw exception
  assert(ctx->pending < ctx->max_ios);
  Buffer *buf = (Buffer*)Data_custom_val(ml_buffer);
  int fd = Int_val(ml_fd);
  uint64_t off = Int64_val(ml_off);
  struct iocb **iocbs = &ctx->iocbs[ctx->pending];
  struct iocb *iocb = iocbs[0];
  intptr_t slot = (intptr_t)iocb->data;

  memset(iocb, 0, sizeof(struct iocb));
  io_prep_pread(iocb, fd, buf->buf, buf->len, off);
  buf->pos = 0;
  iocb->data = (void*)slot;
  Store_field(ml_ctx, slot, ml_fn);
  Store_field(ml_ctx, slot + 1, ml_buffer);

  // FIXME: throw exception
  assert(io_submit(ctx->ctx, 1, iocbs) == 1);
  ++ctx->pending;

  CAMLreturn(Val_unit);
}


/* write: fun ctx fd buf off fn -> ()
external write : context -> Unix.file_descr -> buffer -> int64 -> (unit -> unit) -> unit = "caml_aio_write"
*/
CAMLprim value caml_aio_write(value ml_ctx, value ml_fd, value ml_buffer, value ml_off, value ml_fn) {
  CAMLparam5(ml_ctx, ml_fd, ml_buffer, ml_off, ml_fn);
  fprintf(stderr, "### caml_aio_write()\n");
  Context *ctx = (Context*)Data_custom_val(Field(ml_ctx, 0));
  // FIXME: throw exception
  assert(ctx->pending < ctx->max_ios);
  Buffer *buf = (Buffer*)Data_custom_val(ml_buffer);
  int fd = Int_val(ml_fd);
  uint64_t off = Int64_val(ml_off);
  struct iocb **iocbs = &ctx->iocbs[ctx->pending];
  struct iocb *iocb = iocbs[0];
  intptr_t slot = (intptr_t)iocb->data;

  memset(iocb, 0, sizeof(struct iocb));
  io_prep_pwrite(iocb, fd, buf->buf, buf->len, off);
  buf->pos = 0;
  iocb->data = (void*)slot;

  Store_field(ml_ctx, slot, ml_fn);
  Store_field(ml_ctx, slot + 1, ml_buffer);

  // FIXME: throw exception
  assert(io_submit(ctx->ctx, 1, iocbs) == 1);
  ++ctx->pending;

  CAMLreturn(Val_unit);
}

/* run: fun ctx -> ()
external run : context -> unit = "caml_aio_run"
*/
CAMLprim value caml_aio_run(value ml_ctx) {
  CAMLparam1(ml_ctx);
  CAMLlocal2(ml_fn, ml_buf);
  static value * call_result  = NULL;
  static value * call_error   = NULL;
  static value * call_partial = NULL;
  fprintf(stderr, "### caml_aio_run()\n");
  Context *ctx = (Context*)Data_custom_val(Field(ml_ctx, 0));
  Buffer *buf;

  while(ctx->pending > 0) {
    struct io_event events[ctx->pending];
    struct io_event *ep;
    int n;

    // Wait for at least one event
    n = io_getevents(ctx->ctx, 1, ctx->pending, events, NULL);
    fprintf(stderr, "### caml_aio_run(): n = %d\n", n);
    if (n <= 0) break;

    // process callbacks
    for(ep = events; n-- > 0; ep++) {
      struct iocb *iocb = ep->obj;
      intptr_t slot = (intptr_t)iocb->data;
      fprintf(stderr, "### caml_aio_run(): slot = %"PRIdPTR"\n", slot);

      // Get callback and buffer
      ml_fn = Field(ml_ctx, slot);
      ml_buf = Field(ml_ctx, slot + 1);
      buf = (Buffer*)Data_custom_val(ml_buf);

      // Remove callback and buffer and free iocb
      --ctx->pending;
      Store_field(ml_ctx, slot, Val_unit);
      Store_field(ml_ctx, slot + 1, Val_unit);
      ctx->iocbs[ctx->pending] = iocb;

      // Execute callback
      if (ep->res2 != 0) {
	  if (call_error == NULL) {
	      /* First time around, look up by name */
	      call_error = caml_named_value("caml_aio_call_error");
	  }
	  caml_callback2(*call_error, ml_fn, Val_int(ep->res2));
      } else if (ep->res != buf->len) {
	  if (call_partial == NULL) {
	      /* First time around, look up by name */
	      call_partial = caml_named_value("caml_aio_call_partial");
	  }
	  caml_callback3(*call_partial, ml_fn, ml_buf, Val_int(ep->res));
      } else {
	  if (call_result == NULL) {
	      /* First time around, look up by name */
	      call_result = caml_named_value("caml_aio_call_result");
	  }
	  caml_callback2(*call_result, ml_fn, ml_buf);
      }
    }
  }
  fprintf(stderr, "### caml_aio_run(): done\n");
  CAMLreturn(Val_unit);
}

/* buffer: fun size -> buffer
external buffer: int -> buffer = "caml_aio_buffer"
*/
CAMLprim value caml_aio_buffer(value ml_len) {
  CAMLparam1(ml_len);
  fprintf(stderr, "### caml_aio_buffer()\n");

  value ml_buf = caml_alloc_custom(&caml_aio_buffer_ops, sizeof(Buffer), BLOCKSIZE, 10 * BLOCKSIZE);
  Buffer *buf = (Buffer*)Data_custom_val(ml_buf);
  int len = Int_val(ml_len);
  memset(buf, 0, sizeof(Buffer));
  buf->len = len;
  if (len < PAGESIZE) len = PAGESIZE;

  // FIXME: throw exception
  assert(posix_memalign((void*)&buf->buf, len, PAGESIZE) == 0);


  CAMLreturn(ml_buf);
}

/* get_string: fun buffer len -> string
external get_string: buffer -> int -> string = "caml_aio_get_string"
*/
CAMLprim value caml_aio_get_string(value ml_buffer, value ml_len) {
  CAMLparam2(ml_buffer, ml_len);
  CAMLlocal1(ml_str);
  fprintf(stderr, "### caml_aio_get_string()\n");
  Buffer *buf = (Buffer*)Data_custom_val(ml_buffer);
  size_t len = Int_val(ml_len);
  ml_str = caml_alloc_string(len);

  // FIXME: throw exception
  assert(buf->pos + len <= buf->len);

  memcpy(String_val(ml_str), &buf->buf[buf->pos], len);
  buf->pos += len;

  CAMLreturn(ml_str);
}

/* put_string: fun buffer string -> unit
external put_string: buffer -> string -> unit = "caml_aio_put_string"
*/
CAMLprim value caml_aio_put_string(value ml_buffer, value ml_str) {
  CAMLparam2(ml_buffer, ml_str);
  fprintf(stderr, "### caml_aio_put_string()\n");
  Buffer *buf = (Buffer*)Data_custom_val(ml_buffer);
  size_t len = string_length(ml_str);

  // FIXME: throw exception
  assert(buf->pos + len <= buf->len);

  memcpy(&buf->buf[buf->pos], String_val(ml_str), len);
  buf->pos += len;

  CAMLreturn(Val_unit);
}

/* rewind: fun buffer -> unit
external rewind: buffer -> unit = "caml_aio_rewind"
*/
CAMLprim value caml_aio_rewind(value ml_buffer) {
  CAMLparam1(ml_buffer);
  fprintf(stderr, "### caml_aio_rewind()\n");
  Buffer *buf = (Buffer*)Data_custom_val(ml_buffer);
  buf->pos = 0;

  CAMLreturn(Val_unit);
}
